(1, 2, 3)[::0]
"""
TRACEBACK:
Traceback (most recent call last):
  File "slice__step_zero_tuple.py", line 1, in <module>
    (1, 2, 3)[::0]
    ~~~~~~~~~~~~~~
ValueError: slice step cannot be zero
"""
