a, b, c = (1, 2)
"""
TRACEBACK:
Traceback (most recent call last):
  File "unpack__not_enough.py", line 1, in <module>
    a, b, c = (1, 2)
    ~~~~~~~
ValueError: not enough values to unpack (expected 3, got 2)
"""
