float("it's")
"""
TRACEBACK:
Traceback (most recent call last):
  File "type__float_repr_single_quote.py", line 1, in <module>
    float("it's")
    ~~~~~~~~~~~~~
ValueError: could not convert string to float: "it's"
"""
