[1, 2, 3].index(4)
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__index_not_found.py", line 1, in <module>
    [1, 2, 3].index(4)
    ~~~~~~~~~~~~~~~~~~
ValueError: list.index(x): x not in list
"""
