','.join([1, 2])
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__join_non_string.py", line 1, in <module>
    ','.join([1, 2])
    ~~~~~~~~~~~~~~~~
TypeError: sequence item 0: expected str instance, int found
"""
