x = len
x.nonexistent
"""
TRACEBACK:
Traceback (most recent call last):
  File "type__builtin_attr_error.py", line 2, in <module>
    x.nonexistent
AttributeError: 'builtin_function_or_method' object has no attribute 'nonexistent'
"""
