','.join()
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__join_no_args.py", line 1, in <module>
    ','.join()
    ~~~~~~~~~~
TypeError: str.join() takes exactly one argument (0 given)
"""
