range([1])
"""
TRACEBACK:
Traceback (most recent call last):
  File "traceback__range_as_int.py", line 1, in <module>
    range([1])
    ~~~~~~~~~~
TypeError: 'list' object cannot be interpreted as an integer
"""
