'hello'.index('x')
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__index_not_found.py", line 1, in <module>
    'hello'.index('x')
    ~~~~~~~~~~~~~~~~~~
ValueError: substring not found
"""
