'hello'.rsplit('')
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__rsplit_empty_sep.py", line 1, in <module>
    'hello'.rsplit('')
    ~~~~~~~~~~~~~~~~~~
ValueError: empty separator
"""
